@@ -4,6 +4,7 @@ import (
 	"bufio"
 	"bytes"
 	"debug/macho"
+	"encoding/hex"
 	"encoding/json"
 	"encoding/xml"
 	"errors"
@@ -17,7 +18,6 @@ import (
 
 	"github.com/go-delve/delve/pkg/logflags"
 	"github.com/go-delve/delve/pkg/proc"
-	"github.com/sirupsen/logrus"
 )
 
 type gdbConn struct {
@@ -47,9 +47,10 @@ type gdbConn struct {
 	goarch                string
 	goos                  string
 
-	useXcmd bool // forces writeMemory to use the 'X' command
+	useXcmd       bool // forces writeMemory to use the 'X' command
+	newRRCmdStyle bool // forces qRRCmd to use the post-5.8.0 style always
 
-	log *logrus.Entry
+	log logflags.Logger
 }
 
 var ErrTooManyAttempts = errors.New("too many transmit attempts")
@@ -81,7 +82,7 @@ func isProtocolErrorUnsupported(err error) bool {
 	return gdberr.code == ""
 }
 
-// GdbMalformedThreadIDError is returned when a the stub responds with a
+// GdbMalformedThreadIDError is returned when the stub responds with a
 // thread ID that does not conform with the Gdb Remote Serial Protocol
 // specification.
 type GdbMalformedThreadIDError struct {
@@ -201,7 +202,7 @@ func (conn *gdbConn) qSupported(multiprocess bool) (features map[string]bool, er
 	resp := strings.Split(string(respBuf), ";")
 	features = make(map[string]bool)
 	for _, stubfeature := range resp {
-		if len(stubfeature) <= 0 {
+		if len(stubfeature) == 0 {
 			continue
 		} else if equal := strings.Index(stubfeature, "="); equal >= 0 {
 			if stubfeature[:equal] == "PacketSize" {
@@ -651,8 +652,20 @@ func (conn *gdbConn) step(th *gdbThread, tu *threadUpdater, ignoreFaultSignal bo
 			if ignoreFaultSignal { // we attempting to read the TLS, a fault here should be ignored
 				return nil
 			}
+			if conn.isDebugserver {
+				// For some reason trying to deliver a signal in vCont step makes
+				// debugserver lockup (no errors, it just gets stuck), store the signal
+				// to deliver it later with the vCont;c
+				th.sig = sig
+				return nil
+			}
 		case _SIGILL, _SIGBUS, _SIGFPE:
-			// propagate these signals to inferior immediately
+			if conn.isDebugserver {
+				// See comment above
+				th.sig = sig
+				return nil
+			}
+			// otherwise propagate these signals to inferior immediately
 		case interruptSignal, breakpointSignal, stopSignal:
 			return nil
 		case childSignal: // stop on debugserver but SIGCHLD on lldb-server/linux
@@ -712,6 +725,17 @@ type stopPacket struct {
 	sig       uint8
 	reason    string
 	watchAddr uint64
+	watchReg  int
+	jstopInfo map[int]stopPacket
+	regs      map[uint64]uint64
+}
+
+type jsonStopPacket struct {
+	Tid    int      `json:"tid"`
+	Signal int      `json:"signal"`
+	Metype int      `json:"metype"`
+	Medata []uint64 `json:"medata"`
+	Reason string   `json:"reason"`
 }
 
 // Mach exception codes used to decode metype/medata keys in stop packets (necessary to support watchpoints with debugserver).
@@ -721,9 +745,10 @@ type stopPacket struct {
 //	https://opensource.apple.com/source/xnu/xnu-4570.1.46/osfmk/mach/i386/exception.h.auto.html
 //	https://opensource.apple.com/source/xnu/xnu-4570.1.46/osfmk/mach/arm/exception.h.auto.html
 const (
-	_EXC_BREAKPOINT   = 6     // mach exception type for hardware breakpoints
-	_EXC_I386_SGL     = 1     // mach exception code for single step on x86, for some reason this is also used for watchpoints
-	_EXC_ARM_DA_DEBUG = 0x102 // mach exception code for debug fault on arm/arm64
+	_EXC_BREAKPOINT     = 6     // mach exception type for hardware breakpoints
+	_EXC_I386_SGL       = 1     // mach exception code for single step on x86, for some reason this is also used for watchpoints
+	_EXC_ARM_DA_DEBUG   = 0x102 // mach exception code for debug fault on arm/arm64
+	_EXC_ARM_BREAKPOINT = 1     // mach exception code for breakpoint on arm/arm64
 )
 
 // executes 'vCont' (continue/step) command
@@ -739,31 +764,53 @@ func (conn *gdbConn) parseStopPacket(resp []byte, threadID string, tu *threadUpd
 			return false, stopPacket{}, fmt.Errorf("malformed stop packet: %s", string(resp))
 		}
 		sp.sig = uint8(sig)
+		sp.watchReg = -1
+		sp.regs = make(map[uint64]uint64)
 
 		if logflags.GdbWire() && gdbWireFullStopPacket {
 			conn.log.Debugf("full stop packet: %s", string(resp))
 		}
 
 		var metype int
-		var medata = make([]uint64, 0, 10)
+		medata := make([]uint64, 0, 10)
 
-		buf := resp[3:]
-		for buf != nil {
-			colon := bytes.Index(buf, []byte{':'})
-			if colon < 0 {
-				break
+		parseMachException := func(sp *stopPacket, metype int, medata []uint64) {
+			// Debugserver does not report watchpoint stops in the standard way preferring
+			// instead the semi-undocumented metype/medata keys.
+			// These values also have different meanings depending on the CPU architecture.
+			switch conn.goarch {
+			case "amd64":
+				if metype == _EXC_BREAKPOINT && len(medata) >= 2 && medata[0] == _EXC_I386_SGL {
+					sp.watchAddr = medata[1] // this should be zero if this is really a single step stop and non-zero for watchpoints
+				}
+			case "arm64":
+				if metype == _EXC_BREAKPOINT && len(medata) >= 2 && (medata[0] == _EXC_ARM_DA_DEBUG || medata[0] == _EXC_ARM_BREAKPOINT) {
+					// The arm64 specification allows for up to 16 debug registers.
+					// The registers are zero indexed, thus a value less than 16 will
+					// be a hardware breakpoint register index.
+					// See: https://developer.arm.com/documentation/102120/0101/Debug-exceptions
+					// TODO(deparker): we can ask debugserver for the number of hardware breakpoints
+					// directly.
+					if medata[1] < 16 {
+						sp.watchReg = int(medata[1])
+					} else {
+						sp.watchAddr = medata[1]
+					}
+				}
 			}
-			key := buf[:colon]
-			buf = buf[colon+1:]
+		}
 
-			semicolon := bytes.Index(buf, []byte{';'})
-			var value []byte
-			if semicolon < 0 {
-				value = buf
-				buf = nil
-			} else {
-				value = buf[:semicolon]
-				buf = buf[semicolon+1:]
+		csp := colonSemicolonParser{buf: resp[3:]}
+		for csp.next() {
+			key, value := csp.key, csp.value
+
+			if reg, err := strconv.ParseUint(string(key), 16, 64); err == nil {
+				// This is a register.
+				v, err := strconv.ParseUint(string(value), 16, 64)
+				if err != nil {
+					return false, stopPacket{}, fmt.Errorf("malformed stop packet: %s", string(resp))
+				}
+				sp.regs[reg] = v
 			}
 
 			switch string(key) {
@@ -788,30 +835,36 @@ func (conn *gdbConn) parseStopPacket(resp []byte, threadID string, tu *threadUpd
 				// mach exception data (debugserver extension)
 				d, _ := strconv.ParseUint(string(value), 16, 64)
 				medata = append(medata, d)
+			case "jstopinfo":
+				jstopinfo, err := hex.DecodeString(string(value))
+				if err != nil {
+					return false, stopPacket{}, fmt.Errorf("malformed stop packet: %s (wrong jstopinfo)", string(resp))
+				}
+				parsedJstopInfo := []jsonStopPacket{}
+				if err := json.Unmarshal(jstopinfo, &parsedJstopInfo); err != nil {
+					return false, stopPacket{}, fmt.Errorf("malformed stop packet: %s (wrong jstopinfo)", string(resp))
+				}
+				sp.jstopInfo = make(map[int]stopPacket)
+				for _, jsp := range parsedJstopInfo {
+					threadID := fmt.Sprintf("%d", jsp.Tid)
+					newStopPacket := stopPacket{
+						threadID: threadID,
+						reason:   jsp.Reason,
+						sig:      uint8(jsp.Signal),
+					}
+					parseMachException(&newStopPacket, jsp.Metype, jsp.Medata)
+					sp.jstopInfo[jsp.Tid] = newStopPacket
+				}
 			}
 		}
 
-		// Debugserver does not report watchpoint stops in the standard way preferring
-		// instead the semi-undocumented metype/medata keys.
-		// These values also have different meanings depending on the CPU architecture.
-		switch conn.goarch {
-		case "amd64":
-			if metype == _EXC_BREAKPOINT && len(medata) >= 2 && medata[0] == _EXC_I386_SGL {
-				sp.watchAddr = medata[1] // this should be zero if this is really a single step stop and non-zero for watchpoints
-			}
-		case "arm64":
-			if metype == _EXC_BREAKPOINT && len(medata) >= 2 && medata[0] == _EXC_ARM_DA_DEBUG {
-				sp.watchAddr = medata[1]
-			}
-		}
+		parseMachException(&sp, metype, medata)
 
 		return false, sp, nil
 
 	case 'W', 'X':
 		// process exited, next two character are exit code
-
 		semicolon := bytes.Index(resp, []byte{';'})
-
 		if semicolon < 0 {
 			semicolon = len(resp)
 		}
@@ -861,27 +914,15 @@ func (conn *gdbConn) queryProcessInfo(pid int) (map[string]string, error) {
 
 	pi := make(map[string]string)
 
-	for len(resp) > 0 {
-		semicolon := bytes.Index(resp, []byte{';'})
-		keyval := resp
-		if semicolon >= 0 {
-			keyval = resp[:semicolon]
-			resp = resp[semicolon+1:]
-		}
-
-		colon := bytes.Index(keyval, []byte{':'})
-		if colon < 0 {
-			continue
-		}
-
-		key := string(keyval[:colon])
-		value := string(keyval[colon+1:])
+	csp := colonSemicolonParser{buf: resp}
+	for csp.next() {
+		key, value := string(csp.key), string(csp.value)
 
 		switch key {
 		case "name":
 			name := make([]byte, len(value)/2)
 			for i := 0; i < len(value); i += 2 {
-				n, _ := strconv.ParseUint(string(value[i:i+2]), 16, 8)
+				n, _ := strconv.ParseUint(value[i:i+2], 16, 8)
 				name[i/2] = byte(n)
 			}
 			pi[key] = string(name)
@@ -945,7 +986,10 @@ func (conn *gdbConn) queryThreads(first bool) (threads []string, err error) {
 }
 
 func (conn *gdbConn) selectThread(kind byte, threadID string, context string) error {
-	if conn.threadSuffixSupported {
+	if conn.threadSuffixSupported && kind != 'c' {
+		// kind == 'c' is still allowed because 'rr' is weird about it's support
+		// for thread suffixes and the specification for it doesn't really say how
+		// it should be used with packets 'bc' and 'bs'.
 		panic("selectThread when thread suffix is supported")
 	}
 	conn.outbuf.Reset()
@@ -1149,9 +1193,15 @@ func (conn *gdbConn) qRRCmd(args ...string) (string, error) {
 	}
 	conn.outbuf.Reset()
 	fmt.Fprint(&conn.outbuf, "$qRRCmd")
-	for _, arg := range args {
+	for i, arg := range args {
 		fmt.Fprint(&conn.outbuf, ":")
-		writeAsciiBytes(&conn.outbuf, []byte(arg))
+		if i == 0 && (conn.newRRCmdStyle || conn.threadSuffixSupported) {
+			// newer versions of RR require the command to be followed by a thread id
+			// and the command name to be unescaped.
+			fmt.Fprintf(&conn.outbuf, "%s:-1", arg)
+		} else {
+			writeAsciiBytes(&conn.outbuf, []byte(arg))
+		}
 	}
 	resp, err := conn.exec(conn.outbuf.Bytes(), "qRRCmd")
 	if err != nil {
@@ -1224,24 +1274,9 @@ func (conn *gdbConn) memoryRegionInfo(addr uint64) (*memoryRegionInfo, error) {
 
 	mri := &memoryRegionInfo{}
 
-	buf := resp
-	for len(buf) > 0 {
-		colon := bytes.Index(buf, []byte{':'})
-		if colon < 0 {
-			break
-		}
-		key := buf[:colon]
-		buf = buf[colon+1:]
-
-		semicolon := bytes.Index(buf, []byte{';'})
-		var value []byte
-		if semicolon < 0 {
-			value = buf
-			buf = nil
-		} else {
-			value = buf[:semicolon]
-			buf = buf[semicolon+1:]
-		}
+	csp := colonSemicolonParser{buf: resp}
+	for csp.next() {
+		key, value := csp.key, csp.value
 
 		switch string(key) {
 		case "start":
@@ -1287,7 +1322,7 @@ func (conn *gdbConn) exec(cmd []byte, context string) ([]byte, error) {
 	return conn.recv(cmd, context, false)
 }
 
-var hexdigit = []byte{'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'}
+const hexdigit = "0123456789abcdef"
 
 func (conn *gdbConn) send(cmd []byte) error {
 	if len(cmd) == 0 || cmd[0] != '$' {
@@ -1345,9 +1380,11 @@ func (conn *gdbConn) recv(cmd []byte, context string, binary bool) (resp []byte,
 		if logflags.GdbWire() {
 			out := resp
 			partial := false
-			if idx := bytes.Index(out, []byte{'\n'}); idx >= 0 && !binary {
-				out = resp[:idx]
-				partial = true
+			if !binary {
+				if idx := bytes.Index(out, []byte{'\n'}); idx >= 0 {
+					out = resp[:idx]
+					partial = true
+				}
 			}
 			if len(out) > gdbWireMaxLen {
 				out = out[:gdbWireMaxLen]
@@ -1536,3 +1573,33 @@ func checksum(packet []byte) (sum uint8) {
 	}
 	return sum
 }
+
+// colonSemicolonParser parses a string in the form:
+//
+//	key1:value1;key2:value2;...;keyN:valueN
+type colonSemicolonParser struct {
+	buf        []byte
+	key, value []byte
+}
+
+func (csp *colonSemicolonParser) next() bool {
+	if len(csp.buf) == 0 {
+		return false
+	}
+	colon := bytes.IndexByte(csp.buf, ':')
+	if colon < 0 {
+		return false
+	}
+	csp.key = csp.buf[:colon]
+	csp.buf = csp.buf[colon+1:]
+
+	semicolon := bytes.IndexByte(csp.buf, ';')
+	if semicolon < 0 {
+		csp.value = csp.buf
+		csp.buf = nil
+	} else {
+		csp.value = csp.buf[:semicolon]
+		csp.buf = csp.buf[semicolon+1:]
+	}
+	return true
+}
