@@ -87,7 +87,7 @@ func RuntimeTypeToDIE(_type *Variable, dataAddr uint64, mds []ModuleData) (typ g
 
 	// go 1.11 implementation: use extended attribute in debug_info
 
-	md := findModuleDataForType(bi, mds, _type.Addr, _type.mem)
+	md := findModuleDataForType(mds, _type.Addr)
 	if md != nil {
 		so := bi.moduleDataToImage(md)
 		if so != nil {
@@ -131,15 +131,9 @@ func resolveParametricType(bi *BinaryInfo, mem MemoryReadWriter, t godwarf.Type,
 	}
 	_type := newVariable("", rtypeAddr, runtimeType, bi, mem)
 
-	var mds []ModuleData
-	if bi.moduleDataCache != nil {
-		mds = bi.moduleDataCache
-	} else {
-		mds, err = LoadModuleData(bi, _type.mem)
-		if err != nil {
-			return ptyp.TypedefType.Type, fmt.Errorf("error loading module data: %v", err)
-		}
-		bi.moduleDataCache = mds
+	mds, err := bi.getModuleData(_type.mem)
+	if err != nil {
+		return ptyp.TypedefType.Type, err
 	}
 
 	typ, _, err := RuntimeTypeToDIE(_type, 0, mds)
@@ -197,3 +191,22 @@ func dwarfToRuntimeType(bi *BinaryInfo, mem MemoryReadWriter, typ godwarf.Type)
 	typeKind, _ = constant.Uint64Val(kindv.Value)
 	return typeAddr, typeKind, true, nil
 }
+
+func dieToRuntimeType(bi *BinaryInfo, mem MemoryReadWriter, typ godwarf.Type) (uint64, error) {
+	typc := typ.Common()
+	so := bi.Images[typc.Index]
+	mds, err := bi.getModuleData(mem)
+	if err != nil {
+		return 0, err
+	}
+	md := findModuleDataForImage(mds, so)
+	if md == nil {
+		return 0, fmt.Errorf("could not allocate type %s", typ.String())
+	}
+	for off, rtdie := range so.runtimeTypeToDIE {
+		if rtdie.offset == typc.Offset {
+			return md.types + off, nil
+		}
+	}
+	return 0, fmt.Errorf("could not allocate type %s", typ.String())
+}
