@@ -4,7 +4,6 @@ import (
 	"bytes"
 	"flag"
 	"fmt"
-	"io/ioutil"
 	"net"
 	"net/http"
 	"os"
@@ -26,6 +25,7 @@ import (
 	"github.com/go-delve/delve/service/debugger"
 	"github.com/go-delve/delve/service/rpc2"
 	"github.com/go-delve/delve/service/rpccommon"
+	"github.com/go-delve/liner"
 )
 
 var testBackend, buildMode string
@@ -42,7 +42,7 @@ func TestMain(m *testing.M) {
 		os.Exit(1)
 	}
 	logflags.Setup(logConf != "", logConf, "")
-	os.Exit(test.RunTestsWithFixtures(m))
+	test.RunTestsWithFixtures(m)
 }
 
 type FakeTerminal struct {
@@ -54,7 +54,7 @@ const logCommandOutput = false
 
 func (ft *FakeTerminal) Exec(cmdstr string) (outstr string, err error) {
 	var buf bytes.Buffer
-	ft.Term.stdout.w = &buf
+	ft.Term.stdout.pw.w = &buf
 	ft.Term.starlarkEnv.Redirect(ft.Term.stdout)
 	err = ft.cmds.Call(cmdstr, ft.Term)
 	outstr = buf.String()
@@ -67,7 +67,7 @@ func (ft *FakeTerminal) Exec(cmdstr string) (outstr string, err error) {
 
 func (ft *FakeTerminal) ExecStarlark(starlarkProgram string) (outstr string, err error) {
 	var buf bytes.Buffer
-	ft.Term.stdout.w = &buf
+	ft.Term.stdout.pw.w = &buf
 	ft.Term.starlarkEnv.Redirect(ft.Term.stdout)
 	_, err = ft.Term.starlarkEnv.Execute("<stdin>", starlarkProgram, "main", nil)
 	outstr = buf.String()
@@ -79,6 +79,7 @@ func (ft *FakeTerminal) ExecStarlark(starlarkProgram string) (outstr string, err
 }
 
 func (ft *FakeTerminal) MustExec(cmdstr string) string {
+	ft.t.Helper()
 	outstr, err := ft.Exec(cmdstr)
 	if err != nil {
 		ft.t.Errorf("output of %q: %q", cmdstr, outstr)
@@ -121,7 +122,7 @@ func withTestTerminalBuildFlags(name string, t testing.TB, buildFlags test.Build
 	if testBackend == "rr" {
 		test.MustHaveRecordingAllowed(t)
 	}
-	os.Setenv("TERM", "dumb")
+	t.Setenv("TERM", "dumb")
 	listener, err := net.Listen("tcp", "127.0.0.1:0")
 	if err != nil {
 		t.Fatalf("couldn't start listener: %s\n", err)
@@ -132,7 +133,7 @@ func withTestTerminalBuildFlags(name string, t testing.TB, buildFlags test.Build
 	}
 	server := rpccommon.NewServer(&service.Config{
 		Listener:    listener,
-		ProcessArgs: []string{test.BuildFixture(name, buildFlags).Path},
+		ProcessArgs: []string{test.BuildFixture(t, name, buildFlags).Path},
 		Debugger: debugger.Config{
 			Backend: testBackend,
 		},
@@ -155,7 +156,7 @@ func withTestTerminalBuildFlags(name string, t testing.TB, buildFlags test.Build
 func TestCommandDefault(t *testing.T) {
 	var (
 		cmds = Commands{}
-		cmd  = cmds.Find("non-existant-command", noPrefix).cmdFn
+		cmd  = cmds.Find("non-existent-command", noPrefix).cmdFn
 	)
 
 	err := cmd(nil, callContext{}, "")
@@ -304,6 +305,9 @@ func TestExitStatus(t *testing.T) {
 }
 
 func TestScopePrefix(t *testing.T) {
+	if runtime.GOARCH == "ppc64le" && buildMode == "pie" {
+		t.Skip("pie mode broken on ppc64le")
+	}
 	const goroutinesLinePrefix = "  Goroutine "
 	const goroutinesCurLinePrefix = "* Goroutine "
 	test.AllowRecording(t)
@@ -456,9 +460,6 @@ func TestScopePrefix(t *testing.T) {
 }
 
 func TestOnPrefix(t *testing.T) {
-	if runtime.GOOS == "freebsd" {
-		t.Skip("test is not valid on FreeBSD")
-	}
 	const prefix = "\ti: "
 	test.AllowRecording(t)
 	lenient := false
@@ -520,9 +521,6 @@ func TestNoVars(t *testing.T) {
 }
 
 func TestOnPrefixLocals(t *testing.T) {
-	if runtime.GOOS == "freebsd" {
-		t.Skip("test is not valid on FreeBSD")
-	}
 	const prefix = "\ti: "
 	test.AllowRecording(t)
 	withTestTerminal("goroutinestackprog", t, func(term *FakeTerminal) {
@@ -564,19 +562,6 @@ func TestOnPrefixLocals(t *testing.T) {
 	})
 }
 
-func countOccurrences(s, needle string) int {
-	count := 0
-	for {
-		idx := strings.Index(s, needle)
-		if idx < 0 {
-			break
-		}
-		count++
-		s = s[idx+len(needle):]
-	}
-	return count
-}
-
 func listIsAt(t *testing.T, term *FakeTerminal, listcmd string, cur, start, end int) {
 	t.Helper()
 	outstr := term.MustExec(listcmd)
@@ -711,7 +696,10 @@ func TestIssue827(t *testing.T) {
 	withTestTerminal("notify-v2", t, func(term *FakeTerminal) {
 		go func() {
 			time.Sleep(1 * time.Second)
-			http.Get("http://127.0.0.1:8888/test")
+			resp, err := http.Get("http://127.0.0.1:8888/test")
+			if err == nil {
+				resp.Body.Close()
+			}
 			time.Sleep(1 * time.Second)
 			term.client.Halt()
 		}()
@@ -732,37 +720,67 @@ func findCmdName(c *Commands, cmdstr string, prefix cmdPrefix) string {
 	return ""
 }
 
+func assertNoError(t *testing.T, err error, str string) {
+	t.Helper()
+	if err != nil {
+		t.Fatalf("%s: %v", str, err)
+	}
+}
+
+func assertNoErrorConfigureCmd(t *testing.T, term *Term, cmdstr string) {
+	t.Helper()
+	err := configureCmd(term, callContext{}, cmdstr)
+	assertNoError(t, err, fmt.Sprintf("error executing configureCmd(%s)", cmdstr))
+}
+
+func assertSubstitutePath(t *testing.T, sp config.SubstitutePathRules, v ...string) {
+	t.Helper()
+	if len(sp) != len(v)/2 {
+		t.Fatalf("wrong number of substitute path rules (expected: %d): %#v", len(v)/2, sp)
+	}
+	for i := range sp {
+		if sp[i].From != v[i*2] || sp[i].To != v[i*2+1] {
+			t.Fatalf("wrong substitute path rule %#v expected (from: %q to %q)", sp[i], v[i*2], v[i*2+1])
+		}
+	}
+}
+
+func assertDebugInfoDirs(t *testing.T, got []string, tgt ...string) {
+	if len(got) != len(tgt) {
+		t.Fatalf("wrong number of debug info directories (got %d expected %d)", len(got), len(tgt))
+	}
+	for i := range got {
+		if got[i] != tgt[i] {
+			t.Fatalf("debug info directories mismatch got: %v expected: %v", got, tgt)
+		}
+	}
+}
+
 func TestConfig(t *testing.T) {
+	var buf bytes.Buffer
 	var term Term
 	term.conf = &config.Config{}
 	term.cmds = DebugCommands(nil)
+	term.stdout = &transcriptWriter{pw: &pagingWriter{w: &buf}}
 
 	err := configureCmd(&term, callContext{}, "nonexistent-parameter 10")
 	if err == nil {
 		t.Fatalf("expected error executing configureCmd(nonexistent-parameter)")
 	}
 
-	err = configureCmd(&term, callContext{}, "max-string-len 10")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(max-string-len): %v", err)
-	}
+	assertNoErrorConfigureCmd(t, &term, "max-string-len 10")
 	if term.conf.MaxStringLen == nil {
 		t.Fatalf("expected MaxStringLen 10, got nil")
 	}
 	if *term.conf.MaxStringLen != 10 {
 		t.Fatalf("expected MaxStringLen 10, got: %d", *term.conf.MaxStringLen)
 	}
-	err = configureCmd(&term, callContext{}, "show-location-expr   true")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(show-location-expr   true)")
-	}
+	assertNoErrorConfigureCmd(t, &term, "show-location-expr   true")
 	if term.conf.ShowLocationExpr != true {
 		t.Fatalf("expected ShowLocationExpr true, got false")
 	}
-	err = configureCmd(&term, callContext{}, "max-variable-recurse 4")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(max-variable-recurse): %v", err)
-	}
+
+	assertNoErrorConfigureCmd(t, &term, "max-variable-recurse 4")
 	if term.conf.MaxVariableRecurse == nil {
 		t.Fatalf("expected MaxVariableRecurse 4, got nil")
 	}
@@ -770,26 +788,13 @@ func TestConfig(t *testing.T) {
 		t.Fatalf("expected MaxVariableRecurse 4, got: %d", *term.conf.MaxVariableRecurse)
 	}
 
-	err = configureCmd(&term, callContext{}, "substitute-path a b")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(substitute-path a b): %v", err)
-	}
-	if len(term.conf.SubstitutePath) != 1 || (term.conf.SubstitutePath[0] != config.SubstitutePathRule{From: "a", To: "b"}) {
-		t.Fatalf("unexpected SubstitutePathRules after insert %v", term.conf.SubstitutePath)
-	}
+	assertNoErrorConfigureCmd(t, &term, "substitute-path a b")
+	assertSubstitutePath(t, term.conf.SubstitutePath, "a", "b")
 
-	err = configureCmd(&term, callContext{}, "substitute-path a")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(substitute-path a): %v", err)
-	}
-	if len(term.conf.SubstitutePath) != 0 {
-		t.Fatalf("unexpected SubstitutePathRules after delete %v", term.conf.SubstitutePath)
-	}
+	assertNoErrorConfigureCmd(t, &term, "substitute-path a")
+	assertSubstitutePath(t, term.conf.SubstitutePath)
 
-	err = configureCmd(&term, callContext{}, "alias print blah")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(alias print blah): %v", err)
-	}
+	assertNoErrorConfigureCmd(t, &term, "alias print blah")
 	if len(term.conf.Aliases["print"]) != 1 {
 		t.Fatalf("aliases not changed after configure command %v", term.conf.Aliases)
 	}
@@ -797,16 +802,62 @@ func TestConfig(t *testing.T) {
 		t.Fatalf("new alias not found")
 	}
 
-	err = configureCmd(&term, callContext{}, "alias blah")
-	if err != nil {
-		t.Fatalf("error executing configureCmd(alias blah): %v", err)
-	}
+	assertNoErrorConfigureCmd(t, &term, "alias blah")
 	if len(term.conf.Aliases["print"]) != 0 {
 		t.Fatalf("alias not removed after configure command %v", term.conf.Aliases)
 	}
 	if findCmdName(term.cmds, "blah", noPrefix) != "" {
 		t.Fatalf("new alias found after delete")
 	}
+
+	err = configureCmd(&term, callContext{}, "show-location-expr")
+	if err == nil {
+		t.Fatalf("no error form configureCmd(show-location-expr)")
+	}
+	if !term.conf.ShowLocationExpr {
+		t.Fatalf("ShowLocationExpr not set to true")
+	}
+
+	assertNoErrorConfigureCmd(t, &term, "show-location-expr false")
+	if term.conf.ShowLocationExpr {
+		t.Fatalf("ShowLocationExpr set to true")
+	}
+
+	assertNoErrorConfigureCmd(t, &term, "substitute-path a b")
+	assertNoErrorConfigureCmd(t, &term, "substitute-path c d")
+	assertSubstitutePath(t, term.conf.SubstitutePath, "a", "b", "c", "d")
+
+	buf.Reset()
+	assertNoErrorConfigureCmd(t, &term, "substitute-path")
+	t.Logf("current substitute-path: %q", buf.String())
+	if buf.String() != "\"a\" → \"b\"\n\"c\" → \"d\"\n" {
+		t.Fatalf("wrong substitute-path value")
+	}
+
+	assertNoErrorConfigureCmd(t, &term, "substitute-path -clear c")
+	assertSubstitutePath(t, term.conf.SubstitutePath, "a", "b")
+
+	assertNoErrorConfigureCmd(t, &term, "substitute-path -clear")
+	assertSubstitutePath(t, term.conf.SubstitutePath)
+
+	assertNoErrorConfigureCmd(t, &term, "substitute-path \"\" something")
+	assertSubstitutePath(t, term.conf.SubstitutePath, "", "something")
+
+	assertNoErrorConfigureCmd(t, &term, "substitute-path somethingelse \"\"")
+	assertSubstitutePath(t, term.conf.SubstitutePath, "", "something", "somethingelse", "")
+
+	assertDebugInfoDirs(t, term.conf.DebugInfoDirectories)
+
+	assertNoErrorConfigureCmd(t, &term, "debug-info-directories -add a")
+	assertDebugInfoDirs(t, term.conf.DebugInfoDirectories, "a")
+	assertNoErrorConfigureCmd(t, &term, "debug-info-directories -add b")
+	assertDebugInfoDirs(t, term.conf.DebugInfoDirectories, "a", "b")
+	assertNoErrorConfigureCmd(t, &term, "debug-info-directories -add c")
+	assertDebugInfoDirs(t, term.conf.DebugInfoDirectories, "a", "b", "c")
+	assertNoErrorConfigureCmd(t, &term, "debug-info-directories -rm b")
+	assertDebugInfoDirs(t, term.conf.DebugInfoDirectories, "a", "c")
+	assertNoErrorConfigureCmd(t, &term, "debug-info-directories -clear")
+	assertDebugInfoDirs(t, term.conf.DebugInfoDirectories)
 }
 
 func TestIssue1090(t *testing.T) {
@@ -825,6 +876,9 @@ func TestIssue1090(t *testing.T) {
 }
 
 func TestPrintContextParkedGoroutine(t *testing.T) {
+	if runtime.GOARCH == "ppc64le" && buildMode == "pie" {
+		t.Skip("pie mode broken on ppc64le")
+	}
 	withTestTerminal("goroutinestackprog", t, func(term *FakeTerminal) {
 		term.MustExec("break stacktraceme")
 		term.MustExec("continue")
@@ -848,7 +902,7 @@ func TestPrintContextParkedGoroutine(t *testing.T) {
 		frameout := strings.Split(term.MustExec("frame 0"), "\n")
 		t.Logf("frame 0 -> %q", frameout)
 		if strings.Contains(frameout[0], "stacktraceme") {
-			t.Fatal("bad output for `frame 0` command on a parked goorutine")
+			t.Fatal("bad output for `frame 0` command on a parked goroutine")
 		}
 
 		listout := strings.Split(term.MustExec("list"), "\n")
@@ -898,6 +952,9 @@ func TestOptimizationCheck(t *testing.T) {
 }
 
 func TestTruncateStacktrace(t *testing.T) {
+	if runtime.GOARCH == "ppc64le" && buildMode == "pie" {
+		t.Skip("pie mode broken on ppc64le")
+	}
 	const stacktraceTruncatedMessage = "(truncated)"
 	withTestTerminal("stacktraceprog", t, func(term *FakeTerminal) {
 		term.MustExec("break main.stacktraceme")
@@ -918,6 +975,9 @@ func TestTruncateStacktrace(t *testing.T) {
 func TestIssue1493(t *testing.T) {
 	// The 'regs' command without the '-a' option should only return
 	// general purpose registers.
+	if runtime.GOARCH == "ppc64le" {
+		t.Skip("skipping, some registers such as vector registers are currently not loaded")
+	}
 	withTestTerminal("continuetestprog", t, func(term *FakeTerminal) {
 		r := term.MustExec("regs")
 		nr := len(strings.Split(r, "\n"))
@@ -936,6 +996,9 @@ func findStarFile(name string) string {
 }
 
 func TestIssue1598(t *testing.T) {
+	if buildMode == "pie" && runtime.GOARCH == "ppc64le" {
+		t.Skip("Debug function call Test broken in PIE mode")
+	}
 	test.MustSupportFunctionCalls(t, testBackend)
 	withTestTerminal("issue1598", t, func(term *FakeTerminal) {
 		term.MustExec("break issue1598.go:5")
@@ -961,7 +1024,7 @@ func TestExamineMemoryCmd(t *testing.T) {
 			t.Fatalf("could convert %s into int64, err %s", addressStr, err)
 		}
 
-		res := term.MustExec("examinemem  -count 52 -fmt hex " + addressStr)
+		res := term.MustExec("examinemem  -count 51 -fmt hex " + addressStr)
 		t.Logf("the result of examining memory \n%s", res)
 		// check first line
 		firstLine := fmt.Sprintf("%#x:   0x0a   0x0b   0x0c   0x0d   0x0e   0x0f   0x10   0x11", address)
@@ -970,7 +1033,7 @@ func TestExamineMemoryCmd(t *testing.T) {
 		}
 
 		// check last line
-		lastLine := fmt.Sprintf("%#x:   0x3a   0x3b   0x3c   0x00", address+6*8)
+		lastLine := fmt.Sprintf("%#x:   0x3a   0x3b   0x3c", address+6*8)
 		if !strings.Contains(res, lastLine) {
 			t.Fatalf("expected last line: %s", lastLine)
 		}
@@ -1158,6 +1221,38 @@ func TestHitCondBreakpoint(t *testing.T) {
 			t.Fatalf("wrong value of i")
 		}
 	})
+
+	withTestTerminal("condperghitcount", t, func(term *FakeTerminal) {
+		term.MustExec("break bp1 main.main:8")
+		term.MustExec("condition -per-g-hitcount bp1 == 2")
+		listIsAt(t, term, "continue", 16, -1, -1)
+		// first g hit
+		out := term.MustExec("print j")
+		t.Logf("%q", out)
+		if !strings.Contains(out, "2\n") {
+			t.Fatalf("wrong value of j")
+		}
+		listIsAt(t, term, "continue", 16, -1, -1)
+		// second g hit
+		out = term.MustExec("print j")
+		t.Logf("%q", out)
+		if !strings.Contains(out, "2\n") {
+			t.Fatalf("wrong value of j")
+		}
+	})
+}
+
+func TestCondBreakpointWithFrame(t *testing.T) {
+	withTestTerminal("condframe", t, func(term *FakeTerminal) {
+		term.MustExec("break bp1 callme2")
+		term.MustExec("condition bp1 runtime.frame(1).i == 3")
+		term.MustExec("continue")
+		out := term.MustExec("frame 1 print i")
+		t.Logf("%q", out)
+		if !strings.Contains(out, "3\n") {
+			t.Fatalf("wrong value of i")
+		}
+	})
 }
 
 func TestClearCondBreakpoint(t *testing.T) {
@@ -1271,19 +1366,19 @@ func TestTranscript(t *testing.T) {
 	withTestTerminal("math", t, func(term *FakeTerminal) {
 		term.MustExec("break main.main")
 		out := term.MustExec("continue")
-		if !strings.HasPrefix(out, "> main.main()") {
+		if !strings.HasPrefix(out, "> [Breakpoint 1] main.main()") {
 			t.Fatalf("Wrong output for next: <%s>", out)
 		}
-		fh, err := ioutil.TempFile("", "test-transcript-*")
+		fh, err := os.CreateTemp("", "test-transcript-*")
 		if err != nil {
-			t.Fatalf("TempFile: %v", err)
+			t.Fatalf("CreateTemp: %v", err)
 		}
 		name := fh.Name()
 		fh.Close()
 		t.Logf("output to %q", name)
 
 		slurp := func() string {
-			b, err := ioutil.ReadFile(name)
+			b, err := os.ReadFile(name)
 			if err != nil {
 				t.Fatalf("could not read transcript file: %v", err)
 			}
@@ -1311,3 +1406,198 @@ func TestTranscript(t *testing.T) {
 		os.Remove(name)
 	})
 }
+
+func TestDisassPosCmd(t *testing.T) {
+	if runtime.GOARCH == "ppc64le" && buildMode == "pie" {
+		t.Skip("pie mode broken on ppc64le")
+	}
+	withTestTerminal("testvariables2", t, func(term *FakeTerminal) {
+		term.MustExec("continue")
+		out := term.MustExec("step-instruction")
+		t.Logf("%q\n", out)
+		if !strings.Contains(out, "call $runtime.Breakpoint") && !strings.Contains(out, "CALL runtime.Breakpoint(SB)") {
+			t.Errorf("output doesn't look like disassembly")
+		}
+	})
+}
+
+func TestCreateBreakpointByLocExpr(t *testing.T) {
+	withTestTerminal("math", t, func(term *FakeTerminal) {
+		out := term.MustExec("break main.main")
+		position1 := strings.Split(out, " set at ")[1]
+		term.MustExec("continue")
+		term.MustExec("clear 1")
+		out = term.MustExec("break +0")
+		position2 := strings.Split(out, " set at ")[1]
+		if position1 != position2 {
+			t.Fatalf("mismatched positions %q and %q\n", position1, position2)
+		}
+	})
+}
+
+func TestCreateBreakpointWithCondition(t *testing.T) {
+	withTestTerminal("break", t, func(term *FakeTerminal) {
+		term.MustExec("break bp1 main.main:4 if i == 3")
+		out := term.MustExec("breakpoints")
+		if !strings.Contains(out, "Breakpoint bp1") {
+			t.Fatal("incorrect breakpoint name")
+		}
+		listIsAt(t, term, "continue", 7, -1, -1)
+		out = term.MustExec("print i")
+		t.Logf("%q", out)
+		if !strings.Contains(out, "3\n") {
+			t.Fatalf("wrong value of i")
+		}
+	})
+}
+
+func TestCreateBreakpointWithCondition2(t *testing.T) {
+	withTestTerminal("break", t, func(term *FakeTerminal) {
+		term.MustExec("continue main.main:4")
+		term.MustExec("break if i == 3")
+		out := term.MustExec("breakpoints")
+		if strings.Contains(out, "Breakpoint if") {
+			t.Fatal("incorrect breakpoint name, should be ID")
+		}
+		listIsAt(t, term, "continue", 7, -1, -1)
+		out = term.MustExec("print i")
+		t.Logf("%q", out)
+		if !strings.Contains(out, "3\n") {
+			t.Fatalf("wrong value of i")
+		}
+	})
+}
+
+func TestCreateBreakpointWithCondition3(t *testing.T) {
+	withTestTerminal("test if path/main", t, func(term *FakeTerminal) {
+		// We should not attempt to parse this as a condition.
+		term.MustExec(`break _fixtures/test if path/main.go:4`)
+		listIsAt(t, term, "continue", 4, -1, -1)
+	})
+}
+
+func TestRestartBreakpoints(t *testing.T) {
+	// Tests that breakpoints set using just a line number and with a line
+	// offset are preserved after restart. See issue #3423.
+	withTestTerminal("continuetestprog", t, func(term *FakeTerminal) {
+		term.MustExec("break main.main")
+		term.MustExec("continue")
+		term.MustExec("break 9")
+		term.MustExec("break +1")
+		out := term.MustExec("breakpoints")
+		t.Log("breakpoints before:\n", out)
+		term.MustExec("restart")
+		out = term.MustExec("breakpoints")
+		t.Log("breakpoints after:\n", out)
+		bps, err := term.client.ListBreakpoints(false)
+		assertNoError(t, err, "ListBreakpoints")
+		for _, bp := range bps {
+			if bp.ID < 0 {
+				continue
+			}
+			if bp.Addr == 0 {
+				t.Fatalf("breakpoint %d has address 0", bp.ID)
+			}
+		}
+	})
+}
+
+func TestListPackages(t *testing.T) {
+	test.AllowRecording(t)
+	withTestTerminal("goroutinestackprog", t, func(term *FakeTerminal) {
+		out := term.MustExec("packages")
+		t.Logf("> packages\n%s", out)
+		seen := map[string]bool{}
+		for _, p := range strings.Split(strings.TrimSpace(out), "\n") {
+			seen[p] = true
+		}
+		if !seen["main"] || !seen["runtime"] {
+			t.Error("output omits 'main' and 'runtime'")
+		}
+
+		out = term.MustExec("packages runtime")
+		t.Logf("> packages runtime\n%s", out)
+
+		for _, p := range strings.Split(strings.TrimSpace(out), "\n") {
+			if !strings.Contains(p, "runtime") {
+				t.Errorf("output includes unexpected %q", p)
+			}
+			seen[p] = true
+		}
+		if !seen["runtime"] {
+			t.Error("output omits 'runtime'")
+		}
+	})
+}
+
+func TestSubstitutePathAndList(t *testing.T) {
+	// checks that substitute path rules do not remain cached after a -clear.
+	// See issue #3565.
+	if runtime.GOOS == "windows" {
+		t.Skip("test is not valid on windows due to path separators")
+	}
+	withTestTerminal("math", t, func(term *FakeTerminal) {
+		term.MustExec("break main.main")
+		term.MustExec("continue")
+		fixturesDir, _ := filepath.Abs(test.FindFixturesDir())
+		term.MustExec("config substitute-path " + fixturesDir + " /blah")
+		out, _ := term.Exec("list")
+		t.Logf("list output %s", out)
+		if !strings.Contains(out, "/blah/math.go") {
+			t.Fatalf("bad output")
+		}
+		term.MustExec("config substitute-path -clear")
+		term.MustExec("config substitute-path " + fixturesDir + " /blah2")
+		out, _ = term.Exec("list")
+		t.Logf("list output %s", out)
+		if !strings.Contains(out, "/blah2/math.go") {
+			t.Fatalf("bad output")
+		}
+	})
+}
+
+func TestDisplay(t *testing.T) {
+	// Tests that display command works. See issue #3595.
+	type testCase struct{ in, tgt string }
+	withTestTerminal("testvariables2", t, func(term *FakeTerminal) {
+		term.MustExec("continue")
+
+		for _, tc := range []testCase{
+			{"string(byteslice)", `0: string(byteslice) = "tèst"`},
+			{"string(byteslice[1:])", `0: string(byteslice[1:]) = "èst"`},
+			{"%s string(byteslice)", `0: string(byteslice) = tèst`},
+		} {
+			out := term.MustExec("display -a " + tc.in)
+			t.Logf("%q -> %q", tc.in, out)
+			if !strings.Contains(out, tc.tgt) {
+				t.Errorf("wrong output for 'display -a %s':\n\tgot: %q\n\texpected: %q", tc.in, out, tc.tgt)
+			}
+			term.MustExec("display -d 0")
+		}
+	})
+}
+
+func TestBreakPointFailWithCond(t *testing.T) {
+	if runtime.GOOS == "freebsd" || runtime.GOOS == "darwin" {
+		t.Skip("follow exec not implemented")
+	}
+
+	oldYesNo := yesno
+	defer func() { yesno = oldYesNo }()
+	// always answer yes here
+	yesno = func(line *liner.State, question, defaultAnswer string) (bool, error) {
+		return true, nil
+	}
+
+	withTestTerminal("spawn", t, func(term *FakeTerminal) {
+		assertNoError(t, term.client.FollowExec(true, ""), "FollowExec")
+		_, err := term.Exec("break spawnchild.go:11 if i == 1")
+		if err != nil {
+			t.Errorf("expect to set a suspened breakpoint")
+		}
+		bp, _ := term.client.GetBreakpoint(1)
+		if bp.Cond != "i == 1" {
+			t.Errorf("expected condition to be 'i == 1', got %s", bp.Cond)
+		}
+	})
+}
